/* Level 5 theremin
 *  This program plays songs presented by:
        https://dragaosemchama.com/en/2019/02/songs-for-arduino/
   It uses DAC circuit described by Amanda Ghassaei
        https://www.instructables.com/id/Arduino-Audio-Output/
   Implements 6 bit DAC on PORTB (pins 8-13)
   Plays songs with optional tremolo effect
R. Kubichek
*/
#include "pitches.h"        //copied from "toneMelody" in Examples/Digital group.
#define NumTableEntries 100  //Make smaller if long song runs you short of memory
#define SampleRate 16130     //sample rate. This is defined in function: set_interrupts()
volatile boolean tremulo = false;   //Set this to false to turn off tremulo effect
//User can experiment with different values of tremSweep and tremFreq
int tremSweep = 20; //max freq deviation from desiredFreq. Hz (suggest 20)
int tremFreq = 7;   //controls how fast the waa-waa changes. Hz (suggest 7)
volatile unsigned int desiredFreq;
volatile boolean soundOn = false;   //use to turn on/off output to DAC
byte SineTable[NumTableEntries];    //sine table scaled to 8 bits
int SineTabTrem[NumTableEntries];  //scaled for tremulo shift. can be negative
float deltaTab;
#define REST      0                 //apparently used in the song .h files

//                    ---CHOOSE A SONG HERE---

// melody array contains notes followed by the duration:
//    a 4 means a quarter note, 8 an eighteenth , 16 sixteenth, so on
//    !!negative numbers are used to represent dotted notes,
//    so -4 means a dotted quarter note, that is, a quarter plus an eighteenth!!

//Following melody arrays copied from: https://dragaosemchama.com/en/2019/02/songs-for-arduino/
//To select a song, uncomment both a tempo and an include.  Set tempo as desired
//int tempo = 100;
//#include "lullaby.h"          //suggest tempo 80 to 100
//int tempo = 100;
//#include "starwars.h"         //suggest at least tempo = 100
//int tempo = 150;
//#include "merryChristmas.h"   //suggest tempo = 150
int tempo = 100;
#include "silentNight.h"        //suggest tempo = 100
//int tempo = 150;
//#include "lionSleeps.h"       //suggest tempo = 150
//int tempo = 80;
//#include "greensleeves.h"     //suggest tempo=80
//int tempo = 100;
//#include "darth.h"            //tempo = 100 is ok

int notes = sizeof(melody) / sizeof(melody[0]) / 2;
int wholenote = (60000 * 4) / tempo; //duration of whole note in ms

void set_interrupts() {
  //This routine sets DAC up using timer2 interrupt at ~16 kHz.
  //Avoid timer 0 since it is used by millis()and delay. Also, timer 1 is used for pwm
  //Based on amandaghassaei: https://www.instructables.com/Arduino-Timer-Interrupts/
  cli();//disable interrupts
  TCCR2A = 0;// set entire TCCR2A register to 0
  TCCR2B = 0;// same for TCCR2B
  TCNT2  = 0;//initialize counter value to 0
  //          set compare match register for 16khz increments
  OCR2A = 124; // = (16*10^6) / (16000*8) - 1 (must be <256)
  //              actual freq = (16*10^6)/(prescalar*cmr)+1=16*10^6/(8*124)+1 = 16130
  TCCR2A |= (1 << WGM21);   // turn on CTC mode
  TCCR2B |= (1 << CS21);    // Set CS21 bit for 8 prescaler
  TIMSK2 |= (1 << OCIE2A);  // enable timer compare interrupt
  sei();  //enable interrupts
}
void FillSineTable() {  //precompute sine tables to speed things up
  for (int i = 0; i < NumTableEntries; i++) { // fill table values
    float S = sin(i * (float)(2*PI / NumTableEntries));
    SineTable[i] = 127 * S + 127; //scales to 8 bits
    SineTabTrem[i] = tremSweep * S; //scale to tremulo freq shift
  }
}
ISR(TIMER2_COMPA_vect) { //timer 2 interrupt
  /*  Gets a value from the sine table and outputs to the PORT
     Tremulo: optionally, a small sinusoidally changing increment is added to desiredFreq.
        It causes a "waa-waa" sound. The rate of waa-waa is given by tremFreq.
        The max tremulo freq deviation is given by tremSweep. See FillSineTable
  */  
  static unsigned int index = 0;
  static unsigned int tremindx = 0;
  int FreqAdj = 0;
  byte out;

  //tremulo effect
  if (tremulo) {
    tremindx += tremFreq;    // calculate the next phase
    if (tremindx >= SampleRate)   // keep phase between 0-2*pi
      tremindx -= SampleRate;
    unsigned int indxFrq = (tremindx * deltaTab);  //scale pointer to table size
    FreqAdj = SineTabTrem[indxFrq];  //adjust dsrd freq up or down
    if (FreqAdj + desiredFreq < 0) FreqAdj = 0; //in case desiredFreq = 0
  }
  //main freq pointer
  index += desiredFreq + FreqAdj;  // calculate the next phase
  if (index >= SampleRate)  // keep phase between 0-2*pi
    index -= SampleRate;
  int findx = (index * deltaTab);    //scale pointer to table size
  out = SineTable[findx];
  if (soundOn) PORTB = out>>2;  //remove 2 least-significant bits 
  else PORTB = 31; //the "dc" value
}

void setup() {
  Serial.begin(9600);
  for (int i = 8; i < 14; i++) pinMode(i, OUTPUT);  
  set_interrupts();
  FillSineTable();
  deltaTab = (float) NumTableEntries / (float) SampleRate; //used for scaling
}

void loop() {
  int divider, noteDuration;
  // iterate over the notes of the melody.
  // Remember, the array is twice the number of notes (notes + durations)
  for (int thisNote = 0; thisNote < notes * 2; thisNote = thisNote + 2) {
    // calculate the duration of each note
    divider = melody[thisNote + 1];
    if (divider > 0) { // regular note, just proceed
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) { // dotted notes represented by negative durations!!
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5; // increases the duration in half for dotted notes
    }
    soundOn = true;                 //turns on output sound 
    desiredFreq = melody[thisNote];  
    delay(noteDuration);            //hold the note
    soundOn = false;                //pause between notes
    delay(noteDuration * .2);
  }
  delay(1000); //wait a sec before repeating the song
}
